import pytest
import torch
from torch.distributions import MultivariateNormal

from concept_erasure import optimal_linear_shrinkage


@pytest.mark.parametrize(
    "p,n",
    [
        # Test the n < p case
        (32, 16),
        (64, 32),
        (128, 64),
        # And the n > p case
        (4, 64),
        (8, 128),
        (16, 256),
    ],
)
def test_olse_shrinkage(p: int, n: int):
    torch.manual_seed(42)

    # Number of matrices
    N = 1000

    # Generate a random covariance matrix
    A = torch.randn(N, p, p)
    S_true = A @ A.mT / p
    torch.linalg.diagonal(S_true).add_(1e-3)

    # Generate data with this covariance
    mean = torch.zeros(N, p)
    dist = MultivariateNormal(mean, S_true)
    X = dist.sample([n]).movedim(1, 0)
    assert X.shape == (N, n, p)

    # Compute the sample covariance matrix
    X_centered = X - X.mean(dim=0, keepdim=True)
    S_hat = (X_centered.mT @ X_centered) / n

    # Apply shrinkage
    S_olse = optimal_linear_shrinkage(S_hat, n)

    # Check that the Frobenius norm of the difference has decreased
    norm_naive = torch.norm(S_hat - S_true, dim=(-1, -2)).mean()
    norm_olse = torch.norm(S_olse - S_true, dim=(-1, -2)).mean()

    assert norm_olse <= norm_naive



from contextlib import contextmanager
from functools import partial
from typing import Callable

from torch import Tensor, nn
from transformers import PreTrainedModel

from .leace import LeaceEraser
from .utils import assert_type, is_norm_layer, mangle_module_path


class ConceptScrubber:
    """Wrapper for a dictionary mapping module paths to `LeaceEraser` objects."""

    def __init__(self, pre_hook: bool = False):
        super().__init__()

        self.erasers: dict[str, LeaceEraser] = {}
        self.pre_hook = pre_hook

    @contextmanager
    def scrub(self, model):
        """Add hooks to the model which apply the erasers during a forward pass."""

        def scrub_hook(key: str, x: Tensor):
            eraser = assert_type(LeaceEraser, self.erasers[key])
            return eraser(x).type_as(x)

        with self.apply_hook(model, scrub_hook):
            yield self

    @contextmanager
    def apply_hook(
        self,
        model: nn.Module,
        hook_fn: Callable[[str, Tensor], Tensor | None],
    ):
        """Apply a `hook_fn` to each submodule in `model` that we're scrubbing."""

        def post_wrapper(_, __, output, name: str) -> Tensor | None:
            key = mangle_module_path(name)
            return hook_fn(key, output)

        def pre_wrapper(_, inputs, name: str) -> tuple[Tensor | None, ...]:
            x, *extras = inputs
            key = mangle_module_path(name)
            return hook_fn(key, x), *extras

        # Unwrap the base model if necessary
        if isinstance(model, PreTrainedModel):
            model = assert_type(PreTrainedModel, model.base_model)

        handles = [
            (
                mod.register_forward_pre_hook(partial(pre_wrapper, name=name))
                if self.pre_hook
                else mod.register_forward_hook(partial(post_wrapper, name=name))
            )
            for name, mod in model.named_modules()
            if is_norm_layer(mod) and mangle_module_path(name) in self.erasers
        ]
        assert len(handles) == len(self.erasers), "Not all erasers can be applied"

        try:
            yield self
        finally:
            # Make sure to remove the hooks even if an exception is raised
            for handle in handles:
                handle.remove()



from functools import wraps
from typing import Callable


def cached_property(func: Callable) -> property:
    """Decorator that converts a method into a lazily-evaluated cached property"""
    # Create a secret attribute name for the cached property
    attr_name = "_cached_" + func.__name__

    @property
    @wraps(func)
    def _cached_property(self):
        # If the secret attribute doesn't exist, compute the property and set it
        if not hasattr(self, attr_name):
            setattr(self, attr_name, func(self))

        # Otherwise, return the cached property
        return getattr(self, attr_name)

    return _cached_property


def invalidates_cache(dependent_prop_name: str) -> Callable:
    """Invalidates a cached property when the decorated function is called"""
    attr_name = "_cached_" + dependent_prop_name

    # The actual decorator
    def _invalidates_cache(func: Callable) -> Callable:
        # The wrapper function
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # Check if the secret attribute exists; if so delete it so that
            # the cached property is recomputed
            if hasattr(self, attr_name):
                delattr(self, attr_name)

            return func(self, *args, **kwargs)

        return wrapper

    return _invalidates_cache



import torch
from torch import Tensor


def optimal_linear_shrinkage(S_n: Tensor, n: int | Tensor) -> Tensor:
    """Optimal linear shrinkage for a sample covariance matrix or batch thereof.

    The formula is distribution-free and asymptotically optimal in the Frobenius norm
    as the dimensionality and sample size tend to infinity.

    See "On the Strong Convergence of the Optimal Linear Shrinkage Estimator for Large
    Dimensional Covariance Matrix" <https://arxiv.org/abs/1308.2608> for details.

    Args:
        S_n: Sample covariance matrices of shape (*, p, p).
        n: Sample size.
    """
    p = S_n.shape[-1]
    assert n > 1 and S_n.shape[-2:] == (p, p)

    # Sigma0 is actually a free parameter; here we're using an isotropic
    # covariance matrix with the same trace as S_n.
    # TODO: Make this configurable, try using diag(S_n) or something
    eye = torch.eye(p, dtype=S_n.dtype, device=S_n.device).expand_as(S_n)
    trace_S = trace(S_n)
    sigma0 = eye * trace_S / p

    sigma0_norm_sq = sigma0.pow(2).sum(dim=(-2, -1), keepdim=True)
    S_norm_sq = S_n.pow(2).sum(dim=(-2, -1), keepdim=True)

    prod_trace = trace(S_n @ sigma0)
    top = trace_S.pow(2) * sigma0_norm_sq / n
    bottom = S_norm_sq * sigma0_norm_sq - prod_trace**2

    alpha = 1 - top / bottom
    beta = (1 - alpha) * prod_trace / sigma0_norm_sq

    return alpha * S_n + beta * sigma0


def trace(matrices: Tensor) -> Tensor:
    """Version of `torch.trace` that works for batches of matrices."""
    diag = torch.linalg.diagonal(matrices)
    return diag.sum(dim=-1, keepdim=True).unsqueeze(-1)



from contextlib import contextmanager
from functools import partial

import torch
from torch import Tensor, nn
from transformers import PreTrainedModel

from .utils import assert_type, is_norm_layer


@contextmanager
def random_scrub(model: PreTrainedModel, subspace_dim: int):
    """Add hooks to the model which erase a random subspace during `forward`."""
    d = model.config.hidden_size

    # Unwrap the base model if necessary
    if isinstance(model, PreTrainedModel):
        model = assert_type(PreTrainedModel, model.base_model)

    u = torch.empty(d, subspace_dim, device=model.device, dtype=model.dtype)
    nn.init.orthogonal_(u)

    def random_erase(_, __, x: Tensor, u: Tensor) -> Tensor:
        return x - u @ (u.T @ x)

    handles = [
        mod.register_forward_hook(partial(random_erase, u=u))
        for mod in model.modules()
        if is_norm_layer(mod)
    ]

    try:
        yield
    finally:
        # Make sure to remove the hooks even if an exception is raised
        for handle in handles:
            handle.remove()



from dataclasses import dataclass
from typing import Literal

import torch
from torch import Tensor

from .caching import cached_property, invalidates_cache
from .shrinkage import optimal_linear_shrinkage

ErasureMethod = Literal["leace", "orth"]


@dataclass(frozen=True)
class LeaceEraser:
    """Compact LEACE eraser that surgically erases a concept from a representation."""

    proj_left: Tensor
    proj_right: Tensor
    bias: Tensor | None

    @classmethod
    def fit(cls, x: Tensor, z: Tensor, **kwargs) -> "LeaceEraser":
        """Convenience method to fit a LeaceEraser on data and return it."""
        return LeaceFitter.fit(x, z, **kwargs).eraser

    @property
    def P(self) -> Tensor:
        """The projection matrix."""
        eye = torch.eye(
            self.proj_left.shape[0],
            device=self.proj_left.device,
            dtype=self.proj_left.dtype,
        )
        return eye - self.proj_left @ self.proj_right

    def __call__(self, x: Tensor) -> Tensor:
        """Apply the projection to the input tensor."""
        delta = x - self.bias if self.bias is not None else x

        # Ensure we do the matmul in the most efficient order.
        x_ = x - (delta @ self.proj_right.T) @ self.proj_left.T
        return x_.type_as(x)


class LeaceFitter:
    """Fits a LEACE eraser that surgically erases a concept from a representation.

    This class implements Least-squares Concept Erasure (LEACE) from
    https://arxiv.org/abs/2306.03819. You can also use a slightly simpler orthogonal
    projection-based method by setting `method="orth"`.

    This class stores the covariance statistics needed to compute the LEACE eraser.
    This allows the statistics to be updated incrementally.

    Since the LEACE projection matrix is guaranteed to be a rank k - 1 perturbation of
    the identity, we store it implicitly in the d x k matrices `proj_left` and
    `proj_right`. The full matrix is given by `torch.eye(d) - proj_left @ proj_right`.
    """

    mean_x: Tensor
    """Running mean of X."""

    mean_z: Tensor
    """Running mean of Z."""

    sigma_xz_: Tensor
    """Unnormalized cross-covariance matrix X^T Z."""

    sigma_xx_: Tensor | None
    """Unnormalized covariance matrix X^T X."""

    n: Tensor
    """Number of X samples seen so far."""

    @classmethod
    def fit(cls, x: Tensor, z: Tensor, **kwargs) -> "LeaceFitter":
        """Convenience method to fit a LeaceFitter on data and return it."""
        n, d = x.shape
        _, k = z.reshape(n, -1).shape

        fitter = LeaceFitter(d, k, device=x.device, dtype=x.dtype, **kwargs)
        return fitter.update(x, z)

    def __init__(
        self,
        x_dim: int,
        z_dim: int,
        method: ErasureMethod = "leace",
        *,
        affine: bool = True,
        constrain_cov_trace: bool = True,
        device: str | torch.device | None = None,
        dtype: torch.dtype | None = None,
        shrinkage: bool = True,
        svd_tol: float = 0.01,
    ):
        """Initialize a `LeaceFitter`.

        Args:
            x_dim: Dimensionality of the representation.
            z_dim: Dimensionality of the concept.
            method: Type of projection matrix to use.
            affine: Whether to use a bias term to ensure the unconditional mean of the
                features remains the same after erasure.
            constrain_cov_trace: Whether to constrain the trace of the covariance of X
                after erasure to be no greater than before erasure. This is especially
                useful when injecting the scrubbed features back into a model. Without
                this constraint, the norm of the model's hidden states may diverge in
                some cases.
            device: Device to put the statistics on.
            dtype: Data type to use for the statistics.
            shrinkage: Whether to use shrinkage to estimate the covariance matrix of X.
            svd_tol: Singular values under this threshold are truncated, both during
                the phase where we do SVD on the cross-covariance matrix, and at the
                phase where we compute the pseudoinverse of the projected covariance
                matrix. Higher values are more numerically stable and result in less
                damage to the representation, but may leave trace correlations intact.
        """
        super().__init__()

        self.x_dim = x_dim
        self.z_dim = z_dim

        self.affine = affine
        self.constrain_cov_trace = constrain_cov_trace
        self.method = method
        self.shrinkage = shrinkage

        assert svd_tol > 0.0, "`svd_tol` must be positive for numerical stability."
        self.svd_tol = svd_tol

        self.mean_x = torch.zeros(x_dim, device=device, dtype=dtype)
        self.mean_z = torch.zeros(z_dim, device=device, dtype=dtype)

        self.n = torch.tensor(0, device=device, dtype=dtype)
        self.sigma_xz_ = torch.zeros(x_dim, z_dim, device=device, dtype=dtype)
        # self.sigma_zz_ = torch.zeros(z_dim, z_dim, device=device, dtype=dtype)

        if self.method == "leace":
            self.sigma_xx_ = torch.zeros(x_dim, x_dim, device=device, dtype=dtype)
        elif self.method == "orth":
            self.sigma_xx_ = None
        else:
            raise ValueError(f"Unknown projection type {self.method}")

    @torch.no_grad()
    @invalidates_cache("eraser")
    def update(self, x: Tensor, z: Tensor) -> "LeaceFitter":
        """Update the running statistics with a new batch of data."""
        d, c = self.sigma_xz_.shape
        x = x.reshape(-1, d).type_as(self.mean_x)
        n, d2 = x.shape

        assert d == d2, f"Unexpected number of features {d2}"
        self.n += n

        # Welford's online algorithm
        delta_x = x - self.mean_x
        self.mean_x += delta_x.sum(dim=0) / self.n
        delta_x2 = x - self.mean_x

        # Update the covariance matrix of X if needed (for LEACE)
        if self.method == "leace":
            assert self.sigma_xx_ is not None
            self.sigma_xx_.addmm_(delta_x.mT, delta_x2)

        z = z.reshape(n, -1).type_as(x)
        assert z.shape[-1] == c, f"Unexpected number of classes {z.shape[-1]}"

        delta_z = z - self.mean_z
        self.mean_z += delta_z.sum(dim=0) / self.n
        delta_z2 = z - self.mean_z

        # Update the cross-covariance matrix
        self.sigma_xz_.addmm_(delta_x.mT, delta_z2)

        return self

    @cached_property
    def eraser(self) -> LeaceEraser:
        """Erasure function lazily computed given the current statistics."""
        eye = torch.eye(self.x_dim, device=self.mean_x.device, dtype=self.mean_x.dtype)

        # Compute the whitening and unwhitening matrices
        if self.method == "leace":
            sigma = self.sigma_xx
            L, V = torch.linalg.eigh(sigma)

            # Threshold used by torch.linalg.pinv
            mask = L > (L[-1] * sigma.shape[-1] * torch.finfo(L.dtype).eps)

            # Assuming PSD; account for numerical error
            L.clamp_min_(0.0)

            W = V * L.rsqrt().where(mask, 0.0) @ V.mT
            W_inv = V * L.sqrt().where(mask, 0.0) @ V.mT
        else:
            W, W_inv = eye, eye

        u, s, _ = torch.linalg.svd(W @ self.sigma_xz, full_matrices=False)

        # Throw away singular values that are too small
        u *= s > self.svd_tol

        proj_left = W_inv @ u
        proj_right = u.T @ W

        if self.constrain_cov_trace and self.method == "leace":
            P = eye - proj_left @ proj_right

            # Prevent the covariance trace from increasing
            sigma = self.sigma_xx
            old_trace = torch.trace(sigma)
            new_trace = torch.trace(P @ sigma @ P.mT)

            # If applying the projection matrix increases the variance, this might
            # cause instability, especially when erasure is applied multiple times.
            # We regularize toward the orthogonal projection matrix to avoid this.
            if new_trace > old_trace:
                Q = eye - u @ u.T

                # Set up the variables for the quadratic equation
                x = new_trace
                y = 2 * torch.trace(P @ sigma @ Q.mT)
                z = torch.trace(Q @ sigma @ Q.mT)
                w = old_trace

                # Solve for the mixture of P and Q that makes the trace equal to the
                # trace of the original covariance matrix
                discr = torch.sqrt(
                    4 * w * x - 4 * w * y + 4 * w * z - 4 * x * z + y**2
                )
                alpha1 = (-y / 2 + z - discr / 2) / (x - y + z)
                alpha2 = (-y / 2 + z + discr / 2) / (x - y + z)

                # Choose the positive root
                alpha = torch.where(alpha1 > 0, alpha1, alpha2).clamp(0, 1)
                P = alpha * P + (1 - alpha) * Q

                # TODO: Avoid using SVD here
                u, s, vh = torch.linalg.svd(eye - P)
                proj_left = u * s.sqrt()
                proj_right = vh * s.sqrt()

        return LeaceEraser(
            proj_left, proj_right, bias=self.mean_x if self.affine else None
        )

    @property
    def sigma_xx(self) -> Tensor:
        """The covariance matrix of X."""
        assert self.n > 1, "Call update() before accessing sigma_xx"
        assert (
            self.sigma_xx_ is not None
        ), "Covariance statistics are not being tracked for X"

        # Accumulated numerical error may cause this to be slightly non-symmetric
        S_hat = (self.sigma_xx_ + self.sigma_xx_.mT) / 2

        # Apply Random Matrix Theory-based shrinkage
        if self.shrinkage:
            return optimal_linear_shrinkage(S_hat / self.n, self.n)

        # Just apply Bessel's correction
        else:
            return S_hat / (self.n - 1)

    @property
    def sigma_xz(self) -> Tensor:
        """The cross-covariance matrix."""
        assert self.n > 1, "Call update() with labels before accessing sigma_xz"
        return self.sigma_xz_ / (self.n - 1)



from warnings import warn

import torch.nn.functional as F
from datasets import Dataset
from transformers import GPTNeoXForCausalLM, LlamaForCausalLM, PreTrainedModel

from ..concept_scrubber import ConceptScrubber
from ..leace import ErasureMethod
from .llama import patch_attention_llama_, scrub_llama
from .neox import patch_attention_neox_, scrub_neox


def patch_attention_(model: PreTrainedModel):
    """Patch the attention layers of a model to use fast attention kernels."""
    ty = model.config.model_type
    if not hasattr(F, "scaled_dot_product_attention"):
        warn(
            "Fast, memory-efficient attention requires PyTorch >= 2.0.0. "
            "For best performance, please upgrade your version of PyTorch."
        )
        return

    if ty == "gpt_neox":
        patch_attention_neox_(model)
    elif ty == "llama":
        patch_attention_llama_(model)
    else:
        raise NotImplementedError(f"Unsupported model type: {type(model)}")


def scrub(
    model: PreTrainedModel,
    train: Dataset,
    z_column: str | None,
    affine: bool = True,
    batch_size: int = 1,
    method: ErasureMethod = "leace",
    sublayers: bool = True,
) -> tuple[ConceptScrubber | None, float]:
    """Scrub a model to remove the fast attention kernels."""
    if isinstance(model, GPTNeoXForCausalLM):
        return scrub_neox(model, train, z_column, batch_size, method, affine)
    elif isinstance(model, LlamaForCausalLM):
        return scrub_llama(
            model, train, z_column, batch_size, method, sublayers, affine
        )
    else:
        raise NotImplementedError(f"Unsupported model type: {type(model).__name__}")



from .concept_scrubber import ConceptScrubber
from .data import chunk_and_tokenize
from .leace import ErasureMethod, LeaceEraser, LeaceFitter
from .random_scrub import random_scrub
from .shrinkage import optimal_linear_shrinkage
from .utils import assert_type, chunk

__all__ = [
    "assert_type",
    "chunk",
    "chunk_and_tokenize",
    "optimal_linear_shrinkage",
    "random_scrub",
    "ConceptScrubber",
    "LeaceEraser",
    "LeaceFitter",
    "ErasureMethod",
]



from itertools import pairwise, product

import pytest
import torch
import torch.nn.functional as F
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC

from concept_erasure import ErasureMethod, LeaceFitter, optimal_linear_shrinkage


@pytest.mark.parametrize("shrinkage", [False, True])
def test_stats(shrinkage: bool):
    batch_size = 10
    num_batches = 5
    num_classes = 2
    num_features = 3
    N = batch_size * num_batches

    fitter = LeaceFitter(num_features, num_classes, shrinkage=shrinkage)

    # Generate random data
    torch.manual_seed(42)
    x_data = [torch.randn(batch_size, num_features) for _ in range(num_batches)]
    y_data = [
        torch.randint(0, num_classes, (batch_size, num_classes))
        for _ in range(num_batches)
    ]
    projections = []

    # Compute cross-covariance matrix using batched updates
    for x, y in zip(x_data, y_data):
        fitter.update(x, y)
        projections.append(fitter.eraser.P)

    # Make sure the cached eraser is getting invalidated on update() correctly
    for p1, p2 in pairwise(projections):
        assert not torch.allclose(p1, p2)

    # Compute the expected cross-covariance matrix using the whole dataset
    x_all = torch.cat(x_data)
    y_all = torch.cat(y_data)
    mean_x = x_all.mean(dim=0)
    mean_y = y_all.type_as(x_all).mean(dim=0)
    x_centered = x_all - mean_x
    y_centered = y_all - mean_y

    expected_cov = torch.einsum("b...m,b...n->...mn", x_centered, x_centered)
    if shrinkage:
        expected_cov = optimal_linear_shrinkage(
            expected_cov / N, batch_size * num_batches
        )
    else:
        expected_cov /= N - 1

    expected_sigma_xz = torch.einsum("b...m,b...n->...mn", x_centered, y_centered)
    expected_sigma_xz /= N - 1

    torch.testing.assert_close(fitter.sigma_xx, expected_cov)
    torch.testing.assert_close(fitter.sigma_xz, expected_sigma_xz)


# Both `1` and `2` are binary classification problems, but `1` means the labels are
# encoded in a 1D one-hot vector, while `2` means the labels are encoded in an
# n x 2 one-hot matrix.
@pytest.mark.parametrize("num_classes", [1, 2, 3, 5, 10, 20])
def test_projection(num_classes: int):
    n, d = 2048, 128
    num_distinct = max(num_classes, 2)

    X, Y = make_classification(
        n_samples=n,
        n_features=d,
        n_classes=num_distinct,
        n_informative=num_distinct,
        random_state=42,
    )
    X_t = torch.from_numpy(X)
    Y_t = torch.from_numpy(Y)
    if num_classes > 1:
        Y_1h = torch.nn.functional.one_hot(Y_t, num_classes)
    else:
        Y_1h = Y_t

    bools = [False, True]
    methods: list[ErasureMethod] = ["leace", "orth"]
    eps = 2e-9
    for affine, method, shrink in product(bools, methods, bools):
        # Shrinkage not applicable to orthogonal projection
        if method == "orth" and shrink:
            continue

        fitter = LeaceFitter(
            d,
            num_classes,
            affine=affine,
            dtype=X_t.dtype,
            method=method,
            shrinkage=shrink,
        ).update(X_t, Y_1h)

        eraser = fitter.eraser
        X_ = eraser(X_t)

        # Check first-order optimality condition. To account for the nullspace
        # constraint, we use the canonical form for oblique projection matrices,
        # fixing the nullspace and only optimizing over the range space.
        A, s, B_t = torch.linalg.svd(eraser.P)
        A, B = A[:, s > 0.5].requires_grad_(True), B_t[s > 0.5].T

        # See "A matrix representation formula for a nonzero projection operator" in
        # https://en.wikipedia.org/wiki/Projection_(linear_algebra)
        P = A @ torch.inverse(B.T @ A) @ B.T
        x_ = (X_t - fitter.mean_x) @ P.T + fitter.mean_x if affine else X_t @ P.T

        # Define a random positive definite inner product
        L = torch.randn(d, d, dtype=X_t.dtype) / d**0.5
        loss = F.mse_loss(x_ @ L, X_t @ L)
        loss.backward()

        # Should be optimal iff we're using LEACE with the bias term and no shrinkage
        assert (A.grad.norm() < eps) == (affine and method == "leace" and not shrink)

        # Check idempotence
        torch.testing.assert_close(eraser(X_), X_)

        # Check that the rank of the update <= num_classes
        rank = torch.linalg.svdvals(X_t - X_).gt(eps).sum()
        assert rank <= num_classes

        # Check that the unconditional mean is unchanged
        if affine:
            torch.testing.assert_close(X_t.mean(dim=0), X_.mean(dim=0))

        # Compute class means and check that they are equal after the projection
        class_means_ = torch.stack(
            [X_[Y_t == c].mean(dim=0) for c in range(num_distinct)]
        )
        torch.testing.assert_close(class_means_[1:], class_means_[:-1])

        # Sanity check that class means are NOT equal before the projection
        class_means = torch.stack(
            [X_t[Y_t == c].mean(dim=0) for c in range(num_distinct)]
        )
        assert not torch.allclose(class_means[1:], class_means[:-1])

        # Logistic regression should not be able to learn anything.
        null_lr = LogisticRegression(penalty=None, tol=0.0).fit(
            # Weirdly, in order for this to work consistently with solver='lbfgs', we
            # need to center the design matrix first; otherwise the coefficients aren't
            # quite small enough. Other solvers don't have this problem but they're
            # slow. In theory, the centroid of X shouldn't matter to the solution.
            X_.numpy() - X_.numpy().mean(axis=0),
            Y,
        )
        assert abs(null_lr.coef_).max() < eps

        # Sanity check that it DOES learn something before the projection
        real_lr = LogisticRegression(penalty=None, tol=0.0).fit(
            # Do the same centering operation here to be consistent
            X - X.mean(axis=0),
            Y,
        )
        assert abs(real_lr.coef_).max() > 0.1

        # Linear SVM shouldn't be able to learn anything either
        null_svm = LinearSVC(
            # The dual formulation injects random noise into the solution
            dual=False,
            # Unfortunately the intercept is subject to L2 regularization; setting this
            # to a large value largely cancels out the effect. Regularizing the
            # intercept can cause the coefficients to get larger than they should be.
            intercept_scaling=1e6,
            tol=eps,
        ).fit(X_.numpy(), Y)
        assert abs(null_svm.coef_).max() < eps

        # But it should learn something before the projection
        real_svm = LinearSVC(dual=False, intercept_scaling=1e6, tol=eps).fit(X, Y)
        assert abs(real_svm.coef_).max() > 0.1
